#!/usr/bin/perl

# --------------------------------------------------------------------
#
#   grm_glt.pl
#   
#   Input: text-file with grammar 
#          (HTK's EBNF notation [plus additional scores])
#          
#   Output: text-file with the HTK format grammar's lattice (and 
#           optinal grammar samples)
#           
#   Usage: grm_glt.pl [Options] -i InputGram -o OutputLat
#          Options:
#          --no-dic      does not create dictionary
#          --no-smpl     doesn't create samples file (requires dict)
#
# --------------------------------------------------------------------
#   This software is free.
#   This banner notice must not be removed.
#
#   Vitor Manuel Martins Cicouro de Pera
#   hpc@speech-rec-vcp.com
#   24/03/2016
# --------------------------------------------------------------------


#use strict;


use Getopt::Long;

GetOptions(
   'no-dic' => \$nodic,
   'no-smpl' => \$nosmpl,
   'i=s' => \$input_file,
   'o=s' => \$output_file);

# create temp lattice
system("HParse ".$input_file." grm_glt_tmpfile");

# read temp lattice
open (FILE, 'grm_glt_tmpfile');
<FILE>; 
$line=<FILE>;
($nn) = $line =~ /(\d+)/;   # number of nodes
@fo = (0) x $nn;

# Xg inputs
@f2 = (0) x $nn;
while (($line=<FILE>) !~ /J=0/) { 
  @s2 = split(/\s+/, $line);
  ($idx) = @s2[0] =~ /(\d+)/;
  if (@s2[1] =~ /_[+-]*\d+/) {
    (@f2[$idx]) = @s2[1] =~ /([+-]*\d+\.\d+)/;   ### Xg input
  }
}

# compute fanouts
@st = split(/\s+/, $line);
($idx) = @st[1] =~ /(\d+)/;
@fo[$idx]++;
while ($line=<FILE>) { 
  @st = split(/\s+/, $line);
  ($idx) = @st[1] =~ /(\d+)/;
  @fo[$idx]++;
}
close (FILE);

# compute probs
# baseline version (same as -l option in HParse)
# version w/ external weights
for ($i=0;$i<@fo;$i++) {
  if (@fo[$i] > 0) {
    @fo[$i] = -log(@fo[$i]);
  } else {
    @fo[$i] = -100.0;
  }
}

# write output lattice file
open (FILE,'grm_glt_tmpfile');
open (FILEO, '>', $output_file);
while (($line=<FILE>) !~ /J=0/) { 
  $line =~ s/_[+-]*\d+\.\d+//;   ### Xg input
  print FILEO "$line";
}
@st = split(/\s+/, $line);
($idx) = @st[1] =~ /(\d+)/;
($idx2) = @st[2] =~ /(\d+)/;
chomp($line);
printf FILEO "%s  l=%.3f\n", $line, @fo[$idx]+@f2[$idx2];
while ($line=<FILE>) { 
  @st = split(/\s+/, $line);
  ($idx) = @st[1] =~ /(\d+)/;
  ($idx2) = @st[2] =~ /(\d+)/;
  chomp($line);
  printf FILEO "%s  l=%.3f\n", $line, @fo[$idx]+@f2[$idx2];
}
close (FILEO);
close (FILE);

# erase temp lattice file
system("rm grm_glt_tmpfile");

# ***** generate dictionaryFile
if ( $nodic == 0 ) {
  open (FILE, $input_file);
  while ($line=<FILE>) { 
    chomp $line;
    $line =~ s/(\$\w+\b)//g;
    $line =~ s/\=/ /g;
    $line =~ s/\(/ /g;
    $line =~ s/\)/ /g;
    $line =~ s/\[/ /g;
    $line =~ s/\]/ /g;
    $line =~ s/\|/ /g;
    $line =~ s/\;/ /g;
    $line =~ s/\s+/ /g;
    $line =~ s/_[+-]*\d+\.\d+/ /g;   ### Xg input
    @st = split(/\s+/, $line);
    push @text, @st;
  }
  close (FILE);
  
  %unique = ();
  foreach $item (@text) {
    $unique{$item}++;
  }
  @u_text = keys %unique;
  @su_text = sort @u_text;
  
  open (FILE, '>', $input_file."-dic");
  foreach $word  (@su_text) {
    if ($word ne '') { 
      $word_t = $word;
      $word_t =~ s/(\W)//g; 
      #$word_t =~ s/(\d)//g; 
      print FILE "$word [$word_t] $word_t\n"; 
    }
  }
  close (FILE);

  # generate grammar samples
  if ( $nosmpl == 0 ) {
    system("HSGen -n 10000 -s ".$output_file." ".$input_file."-dic > ".$input_file."-smpl");
  }
}

