#!/usr/bin/perl

# --------------------------------------------------------------------
#
#   grm_join.pl
#   
#   Input: text-files with partial grammars (HTk's EBNF notation) and 
#          manag/ info
#          
#   Output: text-file with whole grammar (partial gram refs are kept)
#           
#   Usage: grm_join.pl -l(ist) GrammarsList -o(ut) Grammar
#
# --------------------------------------------------------------------
#   This software is free.
#   This banner notice must not be removed.
#
#   Vitor Manuel Martins Cicouro de Pera
#   hpc@speech-rec-vcp.com
#   22/03/2016
# --------------------------------------------------------------------


#use strict;
#use warnings;

use Getopt::Long;

my @grm_fnames;
my @grm_r;

sub main() {
  GetOptions('l=s' => \$grm_list,'o=s' => \$grm_out);
  read_grm_list($grm_list);
  read_grm_r();
  for (my $i=0; $i<(@grm_fnames-1); $i++) {
    my $r = add_prefx_part($i,$grm_fnames[$i],$grm_out);
    add_prefx_r($i,$r);
  }
  save_grm($grm_out);
  return;
}

# read names of files with the partial grammars
# 
sub read_grm_list() {
  open (FILE, '<', "@_");
  while (<FILE>) { 
    chomp;
    push(@grm_fnames,$_);
  }
  close (FILE);
  return;
}

# read grammar's root
# 
sub read_grm_r() {
  my $fname = $grm_fnames[@grm_fnames-1];
  open (FILE, '<', $fname);
  while (<FILE>) { 
    chomp;
    push(@grm_r,$_);
  }
  close (FILE);
  return;
}

# add prefix to the partial grammars
#
sub add_prefx_part() {
  my ($pfxn) = @_[0];
  my ($fname_part_grm) = @_[1];
  my ($fname_out_grm) = @_[2];
  my @lines;
  open (FILE, '<', $fname_part_grm);
  while (<FILE>) { 
    chomp;
    push(@lines,$_);
  }
  close (FILE);
  my @tor = split(/ /,@lines[@lines-1]);
  my $torw = @tor[0];
  $torw =~ s/\$//;
  $pfxn++;
  #my $pfx = '$'.'g'.$pfxn.'_';
  my $pfx = '$'.'g'.$pfxn;
  if ($pfxn == 1) {
    open (FILE, '>', $fname_out_grm);
  } else {
    open (FILE, '>>', $fname_out_grm);
  }
  for (my $i=0; $i<(@lines+0); $i++) {
    $lines[$i] =~ s/\$/$pfx/ge;
    printf FILE "%s\n", $lines[$i];
  }
  printf FILE "\n";
  close (FILE);
  return $torw;
}

# add prefixes to the root 
#
sub add_prefx_r() {
  my ($pfxn) = @_[0];
  my ($rnt) = @_[1];
  $pfxn++;
  #my $pfx = 'g'.$pfxn.'_'."$rnt";
  my $pfx = 'g'.$pfxn."$rnt";
  for (my $i=0; $i<(@grm_r+0); $i++) {
    $grm_r[$i] =~ s/$rnt/$pfx/ge;
  }
  return;
}

# save grammar
#
sub save_grm() {
  my ($fname_out_grm) = @_[0];
  open (FILE, '>>', $fname_out_grm);
  for (my $i=0; $i<(@grm_r+0); $i++) {
    printf FILE "%s\n", $grm_r[$i];
  }
  close (FILE);
  return;
}

main();

