#!/usr/bin/perl

# --------------------------------------------------------------------
#
#   *************** [NFY] ***************
#
#   grm_xpd.pl
#   
#   Input: text-file w/ grammar (HTk notation [+ additional scores])
#          
#   Output: text-file w/ expanded (non-term symbols w/ score) grammar
#           
#   Usage: grm_xpd.pl -i InputGrammar -o ExpandedGrammar
#
# --------------------------------------------------------------------
#   This software is free.
#   This banner notice must not be removed.
#
#   Vitor Manuel Martins Cicouro de Pera
#   hpc@speech-rec-vcp.com
#   23/03/2016
# --------------------------------------------------------------------


#use strict;


use Getopt::Long;

my @lines;
my @new_lines;

sub main() {
  GetOptions('i=s' => \$input_file,'o=s' => \$output_file);
  read_grm($input_file);
  xpnd_defs();
  xpnd_lstl();
  write_grm($output_file);
}

# read grammar 
#
sub read_grm() {
  open (FILE, '<', "@_");
  while (<FILE>) { 
    chomp;
    $_ =~ s/(\=|\(|\)|\||\[|\]|\;)/ $1 /g;
    $_ =~ tr/ //s;
    push(@lines,$_);
  }
  close (FILE);
  return;
}

# expand (only $w_v tkns) def lines
#
sub xpnd_defs() {
  for (my $i=0; $i<(@lines-1); $i++) {
    my @line = split(/\s*=\s*/,$lines[$i]);
    if (@line+0>0) {
      push my @new_line, $line[0]." = ";
      my @ldef = split(/\s+/, $line[1]);
      foreach my $x (@ldef) {
        if ($x =~ /\$[\w\d]+_/) {
          my $nx = replx($x);
          push @new_line, " ".$nx." ";
        } 
        else {
          push @new_line, " ".$x." ";
        }
      }
      my $snew_line = "@new_line";
      $snew_line =~ tr/ //s;
      push(@new_lines, $snew_line);
    }
    else {
      push(@new_lines, "");
    }
  }
  return;
}

# expand (only $w_v tkns) lst line
#
sub xpnd_lstl() {
  my @line = split(/\s+/,$lines[(@lines-1)]);
  my @new_line;
  foreach my $x (@line) {
    if ($x =~ /\$[\w\d]+_/) {
      my $nx = replx($x);
      push @new_line, " ".$nx." ";
    } 
    else {
      push @new_line, " ".$x." ";
    }
  }
  my $snew_line = "@new_line";
  $snew_line =~ tr/ //s;
  push(@new_lines, $snew_line);
  return;
}

# recursively expands grm tkn 
#
sub replx {
  my ($x) = @_;
  my @xl = split(/_/, $x);
  my $xt = $xl[0];
  my $xv = $xl[1];
  my @nx;
  
  my $i = -1;
  do {
    $i++;
  } while ( index($new_lines[$i], "$xt"." ") == -1 );

  my @line = split(/\s*=\s*/, $new_lines[$i]);
  my @ldef = split(/\s+/, $line[1]);
  
  my @new_line;
  my $ldoit = 1;
  my $inbrck = 0;
  foreach my $x (@ldef) {
    if( $x =~ /^[\w\d]+_/ ) {   # non-term-w-v
      my @xxl = split(/_/, $x);
      my $xxt = $xxl[0];
      my $xxv = $xxl[1];
      if( ( $ldoit == 1 or chkxpnd("@new_line") == 1 ) and $inbrck == 0 ) {
        $xxv = sprintf("%+.1f",$xxv+$xv);
        $ldoit = 0;
      }
      push @new_line, " ".$xxt."_".$xxv;
    }
    elsif( $x =~ /^[\w\d]+/ ) {   # non-term-wo-v
      if( ( $ldoit == 1 or chkxpnd("@new_line") == 1 ) and $inbrck == 0 ) {
        push @new_line, " ".$x."_".$xv." ";
        $ldoit = 0;
      }
      else {
        push @new_line, " ".$x." ";
      }
    }
    elsif( $x =~ /^\$[\w\d]+_/ ) {   # term-w-v
      my @xxl = split(/_/, $x);
      my $xxt = $xxl[0];
      my $xxv = $xxl[1];
      if( ( $ldoit == 1 or chkxpnd("@new_line") == 1 ) and $inbrck == 0 ) {
        $xxv += $xv;
        $ldoit = 0;
      }
      my $nx = replx($xxt."_".$xxv);
      push @new_line, " ".$nx." ";
    }
    elsif( $x =~ /^\$[\w\d]+/ ) {   # term-wo-v
      if( ( $ldoit == 1 or chkxpnd("@new_line") == 1 ) and $inbrck == 0 ) {
        my $nx = replx($x."_".$xv);
        push @new_line, " ".$nx." ";
        $ldoit = 0;
      }
      else {
        push @new_line, " ".$x." ";
      }
    }
    elsif( $x =~ m/(\(|\)|\|)/ ) {   # (|)
      push @new_line, " ".$x." ";
    }
    elsif( $x =~ m/\[/ ) {   # [ 
      $inbrck++;
      push @new_line, " ".$x." ";
    }
    elsif( $x =~ m/\]/ ) {   # ] 
      $inbrck--;
      push @new_line, " ".$x." ";
    }
  }
  my $snew_line = "( "."@new_line"." )";
  $snew_line =~ tr/ //s;

  return $snew_line;
}

# check if branch is to expand_ 
#
sub chkxpnd {
  my ($tline) = @_;
  $tline =~ s/_[+-]\d\.\d//g;
  $tline =~ s/[\w\d]+/t/g;
  $tline =~ s/\$t/t/g;
  $tline =~ tr/ +/ /s;
  do {
    $tline =~ s/(t )+/t /g;
    $tline =~ s/\( [^\(\)]* \)/t/g;
  } while $tline =~ m/[\)]+/;
  if ( $tline =~ m/t[ ]*$/ ) {
    return 0;
  }
  else {
    return 1;
  }
}

# write grammar 
#
sub write_grm() {
  open (FILE, '>', "@_");
  for (my $i=0; $i<(@new_lines+0); $i++) {
    print FILE "@new_lines[$i]\n";
  }
  close (FILE);
  return;
}

main();

