#!/usr/bin/perl

# --------------------------------------------------------------------
#
#   trn_prmp.pl
#   
#   Input: application dictionary (text-file)
#          application prompts (text-file)
#          (additional info to compute the prompts scores (scr_val()))
#
#   Output: selected prompts (text-file)
#           
#   Usage: trn_prmp.pl -d Dictionary -p Prompts -o(ut) SelectedPrompts
#   
# --------------------------------------------------------------------
#   This software is free.
#   This banner notice must not be removed.
#
#   Vitor Manuel Martins Cicouro de Pera
#   hpc@speech-rec-vcp.com
#   03/04/2016
# --------------------------------------------------------------------


#use strict;
#use warnings;

use Getopt::Long;
use Sort::Key::Top 'top';

use constant MIN_PHN => 10;     # min num occurs of sub-word units
use constant MIN_PRMPT => 5;   # min num of prompts

my %dict;
my @prpm;
my @prpm_sw;
my @top_prpm;
my %phn_tt;
my %phn_prpm;


sub main() {
  GetOptions('d=s' => \$dict_file,
             'p=s' => \$prpm_file,
             'o=s' => \$slct_file);
  read_dict($dict_file);
  read_prpm($prpm_file);
  nvtr_dict();
  xpd_prpm();
  srand(time);         # delete when scr_val() criteria set
  bst_prpm();
  if (!ngh_prpm()) {
    slc_prpm();
  }
  save_slct($slct_file);
}

# read dictionary
# 
sub read_dict() {
  open (FILE, '<', "@_");
  while ( $entry=<FILE> ) { 
    ($wrd, $def) = split /:\s*/, $entry, 2;
    @phns = split ' ', $def;
    $dict{$wrd} = [ @phns ];
  } 
  close (FILE);
#  #-------------------
#  # dictionary ok?
#  print "\ndictionary:\n";
#  #-------------------
#  #foreach $entry ( keys %dict ) {
#  #  print "$entry: @{ $dict{$entry} }\n";
#  #}
#  # ... or ...
#  foreach $entry ( keys %dict ) {
#    print "$entry: ";
#    foreach $i ( 0 .. $#{ $dict{$entry} } ) {
#      print " $i = $dict{$entry}[$i]";
#    }
#    print "\n";
#  }
#  #-------------------
  return;
}

# read prompts
#
sub read_prpm() {
  open (FILE, '<', "@_");
  my $i=0;
  while ( $sent=<FILE> ) {
    my @lsent = split(/\s+/, $sent);
    push @prpm, [ @lsent ];
  }
  close (FILE);
#  #-------------------
#  # prompts ok?
#  print "\nprompts:\n";
#  for $i ( 0 .. $#prpm ) {
#    print "[ @{$prpm[$i]} ],\n";
#  }
#  #-------------------
  return;
}

# dictionary sub-word units
# 
sub nvtr_dict() {
  foreach $entry ( keys %dict ) {
    foreach $i ( 0 .. $#{ $dict{$entry} } ) {
      $phn_tt{$dict{$entry}[$i]}++;
    }
  }
#  #-------------------
#  print "\ncounts phones:\n";
#  foreach $phnn ( keys %phn_tt ) {
#    print "$phnn: $phn_tt{$phnn}\t";
#  }
#  #-------------------
  return;
}

# expand words into sub-word units
#
sub xpd_prpm()
{
  for $i ( 0 .. $#prpm ) {
    my @lsent = @{$prpm[$i]};
    my @psent;
    foreach $swu (@lsent) {
      push @psent, @{$dict{$swu}};
    }
    push @prpm_sw, "@psent";
  }
  return;
}

# select prompts
# 
sub bst_prpm() {
  my @prpm_scr;
  for my $i ( 0 .. $#prpm_sw ) {
    my $x = scr_val($prpm_sw[$i]);
    my $xfmt = sprintf ("%04s", $x);
    push(@prpm_scr,$xfmt.$i);
  }
  @top_prpm = top &MIN_PRMPT => @prpm_scr;
  for my $i ( 0 .. $#top_prpm ) {
    $top_prpm[$i] =~ s/^\d\d\d\d//;
  }
  return;
}

################################################
# define criteria (rules) and compute prpm score
# input: expanded prompts (@prpm_sw)
#        ...
# output: the prompt score
sub scr_val() {
  my $dm = @_[0];
  #
  # $val = ...
  #----------------------------------------
  $val = int(1+rand(10));
  #----------------------------------------
  return $val;
}

# check if enough prompts
# 
sub ngh_prpm() {
  %phn_prpm = %phn_tt;
  foreach my $key (keys(%phn_prpm)) {
    $phn_prpm{$key} = 0;
  }
  foreach my $i ( 0 .. $#top_prpm ) {
    my @lphn = split / /, $prpm_sw[$top_prpm[$i]];
    foreach my $j ( 0 .. $#lphn ) {
      $phn_prpm{$lphn[$j]}++;
    }
  }
  foreach my $val (values(%phn_prpm)) {
    if ($val < MIN_PHN) {
      return 0;
    }
  }
  return 1;
}

# select additional prompts
# 
sub slc_prpm() {
  my %phn_mss;
  foreach my $key (keys(%phn_prpm)) {
    $phn_mss{$key} = MIN_PHN-$phn_prpm{$key};
  }
#  #-----------------------------------------------
#  for my $i ( 0 .. $#prpm_sw ) { print "\n$i :\t$prpm_sw[$i]"; }
#  print "\n\n1st-top: ". join(" ",@top_prpm);
#  print "\n\nMIN_PHN = ".&MIN_PHN."\n\n";
#  foreach my $key (keys(%phn_mss)) { print "\t $key  $phn_mss{$key}"; }
#  #-----------------------------------------------
  foreach my $key (keys(%phn_mss)) {
#    #-----------------------------------------------
#    print "\nphn: $key ";
#    #-----------------------------------------------
    if ($phn_mss{$key} > 0) {
#      #-----------------------------------------------
#      print "...";
#      #-----------------------------------------------
      my @cnt;
      for my $i ( 0 .. $#prpm_sw ) {
        my $ifmt = sprintf ("%04s", $i);
        if (grep{$_ eq $i} @top_prpm) {
          push @cnt, "999".$ifmt;
	}
        else {
	  my $nn=0;
	  while ($prpm_sw[$i] =~ /\b$key\b/g) { 
	    $nn++; 
	  }
          my $nnfmt = sprintf ("%03s", 999-$nn);
          push @cnt, $nnfmt.$ifmt;
        }
      } 
      my $ntop = $#prpm_sw-$#top_prpm;
      if ($ntop>MIN_PHN) {
        $ntop=MIN_PHN;
      }
      my @top_p = top $ntop => @cnt;   ### @top_p non-sorted
      for my $i ( 0 .. $#top_p ) {
        $top_p[$i] =~ s/^\d\d\d//;
        $top_p[$i] =~ s/^0{0,3}//;
      }
      my $j=0;
      while ($phn_mss{$key} > 0) {
	if ($j > $#top_p) {
	  print "\nNot enough samples of \"$key\" sub-word unit.\n";
	  last;
	}
        if (!grep{$_ eq $top_p[$j]} @top_prpm) {
          push (@top_prpm,"$top_p[$j]");
          my @lphn_cpt = split / /, $prpm_sw[$top_p[$j]];
          foreach my $k ( 0 .. $#lphn_cpt ) {
            $phn_mss{$lphn_cpt[$k]}--;
          }
	}
	$j++;
      }
    }
  }
#  #-----------------------------------------------
#  print "\n";
#  foreach my $key (keys(%phn_mss)) { print "\t $key  $phn_mss{$key}"; }
#  print "\n\n2nd-top: ". join(" ",@top_prpm) ."\n\n";
#  #-----------------------------------------------
  return;
}

# save selected prompts
# 
sub save_slct() {
  open (FILE, '>', "@_");
  for my $i ( 0 .. $#top_prpm ) {
    print FILE "$top_prpm[$i] :\t";
    print FILE join " ", @{$prpm[$top_prpm[$i]]};
    print FILE "\n";
  }
  close (FILE);
  return;
}

main();

