#!/usr/bin/perl

# --------------------------------------------------------------------
#
#   trn_prmp_st_pdt.pl
#   
#   Input: application dictionary (text-file)
#          application prompts (text-file)
#          application Q-prompts (text-file)
#          (additional info to compute the prompts scores (scr_val()))
#
#   Output: selected Q-prompts (text-file)
#           
#   Usage: trn_prmp_st_pdt.pl -d Dictionary \
#                             -s SelectedPrompts \
#                             -p Q-prompts -o(ut) Q-SelectedPrompts
#   
# --------------------------------------------------------------------
#   This software is free.
#   This banner notice must not be removed.
#
#   Vitor Manuel Martins Cicouro de Pera
#   vitor.cicouro@gmail.com
#   06/12/2016
# --------------------------------------------------------------------


#use strict;
#use warnings;

use Getopt::Long;
use Sort::Key::Top 'topsort';

use constant MIN_PHN => 10;     # min num occurs of sub-word units
use constant MAX_PRMPT => 5;   # min num of prompts

my %dict;
my @prpm;
my @prpm_sw;
my $prpm_in;
my @top_prpm;
my %phn_tt;
my %phn_prpm;


sub main() {
  GetOptions('d=s' => \$dict_file,        # Dictionary
             's=s' => \$slc_prpm_file,    # SelectdPrompts
             'p=s' => \$q_prpm_file,        # Q-prompts
             'o=s' => \$slct_file);       # Q-SelectedPrompts
  read_dict($dict_file);
  read_slc_prpm($slc_prpm_file);
  read_q_prpm($q_prpm_file);
  nvtr_dict();
  xpd_prpm();
  srand(time);         # delete when scr_val() criteria set
  bst_prpm();
  if (!ngh_prpm()) {
    printf "\nadditional prompts ...\n";
    slc_prpm();
    save_slct($slct_file);
  } 
  else { 
    empty_slct($slct_file);
  }
}

# read dictionary
# 
sub read_dict() {
  open (FILE, '<', "@_");
  while ( $entry=<FILE> ) { 
    ($wrd, $def) = split /:\s*/, $entry, 2;
    @phns = split ' ', $def;
    $dict{$wrd} = [ @phns ];
  } 
  close (FILE);
  return;
}

# read selected prompts
#
sub read_slc_prpm() {
  open (FILE, '<', "@_");
  my $i=0;
  while ( $sent=<FILE> ) {
    $sent =~ s/^.*\t//s;
    my @lsent = split(/\s+/, $sent);
    push @prpm, [ @lsent ];
  }
  close (FILE);
  $prpm_in = $#prpm+1;
  if ( $prpm_in < MIN_PRMPT ) {
    printf "\nERROR: not enough non-Q prompts!\n";
  }
  return;
}

# read q-prompts
#
sub read_q_prpm() {
  open (FILE, '<', "@_");
  my $i=0;
  while ( $sent=<FILE> ) {
    my @lsent = split(/\s+/, $sent);
    push @prpm, [ @lsent ];
  }
  close (FILE);
  return;
}

# dictionary sub-word units
# 
sub nvtr_dict() {
  ### clean dict (if not in @prpm ...)
  my @prpm_l;
  for my $i ( 0 .. $#prpm ) {
    push @prpm_l, @{$prpm[$i]};
  }
  #my %seen = ();
  #my @uniqu = grep { ! $seen{$_} ++ } @prpm_l;
  foreach my $kk (keys %dict) {
    if ( ! grep{$_ eq $kk} @prpm_l ) {    ### @uniqu ) {
      delete $dict{$kk};
    }
  }

  foreach $entry ( keys %dict ) {
    foreach $i ( 0 .. $#{ $dict{$entry} } ) {
      $phn_tt{$dict{$entry}[$i]}++;
    }
  }
  return;
}

# expand words into sub-word units
#
sub xpd_prpm()
{
  for $i ( 0 .. $#prpm ) {
    my @lsent = @{$prpm[$i]};
    my @psent;
    foreach $swu (@lsent) {
      push @psent, @{$dict{$swu}};
    }
    push @prpm_sw, "@psent";
  }
  return;
}

# prompts score sorting
# 
sub bst_prpm() {
#
#  my @prpm_scr;
#  for my $i ( 0 .. $#prpm_sw ) {
#    my $x = scr_val($prpm_sw[$i]);
#    my $xfmt = sprintf ("%04s", $x);
#    push(@prpm_scr,$xfmt.$i);
#  }
#  @top_prpm = top &MIN_PRMPT => @prpm_scr;
#  for my $i ( 0 .. $#top_prpm ) {
#    $top_prpm[$i] =~ s/^\d\d\d\d//;
#  }
#
#-------------------
# tmp
  for my $i ( 0 .. $prpm_in-1 ) {
    push @top_prpm, "$i";
  }
#-------------------
  return;
}

################################################
# define criteria (rules) and compute prpm score
# input: expanded prompts (@prpm_sw)
#        ...
# output: the prompt score
#sub scr_val() {
#  my $dm = @_[0];
#  #
#  # $val = ...
#  #----------------------------------------
#  $val = int(1+rand(10));
#  #----------------------------------------
#  return $val;
#}

# check if enough prompts
# 
sub ngh_prpm() {
  #
  # tmp: "fit-dictionary" ( ... $#phn_tt >= $#phn_prpm )
  # 
  %phn_prpm = %phn_tt;
  foreach my $key (keys(%phn_prpm)) {
    $phn_prpm{$key} = 0;
  }
  foreach my $i ( 0 .. $#top_prpm ) {
    my @lphn = split / /, $prpm_sw[$top_prpm[$i]];
    foreach my $j ( 0 .. $#lphn ) {
      $phn_prpm{$lphn[$j]}++;
    }
  }
  foreach my $val (values(%phn_prpm)) {
    if ($val < MIN_PHN) {
      return 0;
    }
  }
  return 1;
}

# select additional prompts
# 
sub slc_prpm() {
  my %phn_mss;
  foreach my $key (keys(%phn_prpm)) {
    $phn_mss{$key} = MIN_PHN-$phn_prpm{$key};
  }
  foreach my $key (keys(%phn_mss)) {
    if ($phn_mss{$key} > 0) {
      my @cnt;
      for my $i ( 0 .. $#prpm_sw ) {
        my $ifmt = sprintf ("%04s", $i);
        if (grep{$_ eq $i} @top_prpm) {
          push @cnt, "999".$ifmt;
	}
        else {
	  my $nn=0;
	  while ($prpm_sw[$i] =~ /\b$key\b/g) { 
	    $nn++; 
	  }
          my $nnfmt = sprintf ("%03s", 999-$nn);
          push @cnt, $nnfmt.$ifmt;
        }
      } 
      my $ntop = $#prpm_sw-$#top_prpm;
      if ($ntop>MAX_PRMPT) {
        $ntop=MAX_PRMPT;
      }
      my @top_p = topsort $ntop => @cnt;
      for my $i ( 0 .. $#top_p ) {
        $top_p[$i] =~ s/^\d\d\d//;
        $top_p[$i] =~ s/^0{0,3}//;
      }
      my $j=0;
      while ($phn_mss{$key} > 0) {
	if ($j > $#top_p) {
	  print "\nNot enough samples of \"$key\" sub-word unit.\n";
	  last;
	}
        if (!grep{$_ eq $top_p[$j]} @top_prpm) {
          push (@top_prpm,"$top_p[$j]");
          my @lphn_cpt = split / /, $prpm_sw[$top_p[$j]];
          foreach my $k ( 0 .. $#lphn_cpt ) {
            $phn_mss{$lphn_cpt[$k]}--;
          }
	}
	$j++;
      }
    }
  }
  return;
}

# save selected prompts
# 
sub save_slct() {
  open (FILE, '>', "@_");
  for my $i ( $prpm_in .. $#top_prpm ) {
    my $ndx = $top_prpm[$i]-$prpm_in;
    print FILE "$ndx :\t";
    print FILE join " ", @{$prpm[$top_prpm[$i]]};
    print FILE "\n";
  }
  close (FILE);
  return;
}

# empty prompts file
# 
sub empty_slct() {
  open (FILE, '>', "@_");
  print FILE "\n";
  close (FILE);
  return;
}

main();

